"""add default requires approval field on tools

Revision ID: c41c87205254
Revises: 068588268b02
Create Date: 2025-08-28 13:17:51.636159

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "c41c87205254"
down_revision: Union[str, None] = "068588268b02"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("tools", sa.Column("default_requires_approval", sa.Boolean(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("tools", "default_requires_approval")
    # ### end Alembic commands ###
