"""add api version to byok providers

Revision ID: ffb17eb241fc
Revises: 5fb8bba2c373
Create Date: 2025-08-12 14:35:26.375985

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "ffb17eb241fc"
down_revision: Union[str, None] = "5fb8bba2c373"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("providers", sa.Column("api_version", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("providers", "api_version")
    # ### end Alembic commands ###
